{$A+,B-,F+,G+,N+,E-,P+,Q-,R-,T-,V+,X+}
{****************************************************************************
 * SoundLib 2 Silent driver - (c) 2001 by Stefan Goehler                    *
 * Use this driver if no SB compatible card has been found, to avoid        *
 * problems with non-initialized playback objects                           *
 * .simulates soundcard by retrieving sound data (if wished, set 'docalls'  *
 * tsilent.init to true)                                                    *
 * .alternatively, call tsilent.init with false - you can do anything with  *
 *  your sound objects, nothing will happen - 'dummy driver'                *
 ****************************************************************************}
{$IFDEF FPC}
{$ASMMODE INTEL}
{$ENDIF}
unit drv_sil;

interface
uses sndlib2;

type
   psilent = ^tsilent;
   tsilent = object(tsounddriver)
     constructor init(docalls : boolean);
     destructor  done;virtual;

     function    getcapabilities(var tcap : tcardinfo) : boolean;virtual;
     function    startplayback(freq,flags : longint) : boolean;virtual;
     procedure   stopplayback;virtual;
     function    getversionstring : string;virtual;

     function    getnumofoutcontrols : word;virtual;
     function    getoutcontroltype(num : word) : word;virtual;
     function    getoutcontrolname(cnt : word) : string;virtual;
     function    getoutcontrolrange(cnt : word) : longint;virtual;
     function    getoutcontrol(cnt : word) : longint;virtual;
     function    setoutcontrol(cnt : word;val : longint) : boolean;virtual;
     procedure   intrcallproc;virtual;
     private
     datap : pointer;
   end;

implementation
{$IFNDEF FPC}
uses dos,winapi,newfrontier,snd_cnv,objects;
{$ELSE}
uses dos,snd_cnv,go32,objects;
{$ENDIF}

var
  calls         : boolean;
  oldexitproc   : pointer;
  fpustate      : pointer;
  cap           : tcardinfo;
  silentinit    : boolean;
  activeblock   : word;
  callt         : psilent;

  {$IFNDEF FPC}
  oldint        : pointer;
  {$ELSE}
  oldint        : tseginfo;
  newint        : tseginfo;
  int_ds        : word; external name '___v2prt0_ds_alias';
  {$ENDIF}

const
  versionstring = 'Silent driver v2.0a3';

procedure exitsilent;far;
begin
  { Restore old interrupt caller }
  if calls then
  {$IFNDEF FPC}
  setintvec($1C,oldint);
  {$ELSE}
  set_pm_interrupt($1C,oldint);
  {$ENDIF}
  exitproc := oldexitproc;
end;

procedure intcaller;interrupt;
var
  h   : byte;

begin
  {$IFNDEF FPC}
  asm
    { Disable interrupts }
    cli

    { Save ALL registers }
    db 66h;pusha

           push   es
           push   ds
    db 0Fh,0A8h {push gs}
    db 0Fh,0A0h {push fs}

    { Save FPU state     }
    cmp    test8087,0
    je     @nofpu
      les    di,fpustate
      fsave  es:[di]
    @nofpu:
  end;
  {$ELSE}
  asm
    { Disable interrupts }
    cli

    { Save ALL registers }
    push   ds
    push   ax
    mov    ax,cs:[int_ds]
    mov    ds,ax

    push   es
    push   gs
    push   fs

    pushad

    { Save FPU state     }
    cmp    test8087,0
    je     @nofpu
      mov    edi,fpustate
      fsave  ds:[edi]
    @nofpu:
  end;
  {$ENDIF}
  if assigned(callt) then callt^.intrcallproc;
  activeblock := (activeblock+1) and 1;
  {$IFNDEF FPC}
  asm
    { Restore FPU state    }
    cmp    test8087,0
    je     @nofpu
      les    di,fpustate
      frstor es:[di]
    @nofpu:

    { Reload ALL registers }
    db 0Fh,0A1h {pop fs}
    db 0Fh,0A9h {pop gs}
           pop   ds
           pop   es

    db 66h;popa

    { Enable interrupts    }
    sti
  end;
  {$ELSE}
  asm
    { Restore FPU state    }
    cmp    test8087,0
    je     @nofpu
      mov    edi,fpustate
      frstor ds:[edi]
    @nofpu:

    { Reload ALL registers }
    popad

    pop   fs
    pop   gs
    pop   es

    pop   ax
    pop   ds

    { Enable interrupts    }
    sti
  end;
  {$ENDIF}
end;

constructor tsilent.init(docalls : boolean);
begin
  tobject.init;
  { Exit if already another instance active }
  if (silentinit) then fail;
  if (docalls) then
  begin
    {$IFNDEF FPC}
    getintvec($1C,oldint);
    setintvec($1C,addr(intcaller));
    {$ELSE}
    newint.offset  := addr(intcaller);
    newint.segment := get_cs;
    get_pm_interrupt($1C,oldint);
    set_pm_interrupt($1C,newint);
    {$ENDIF}
  end;
  calls := docalls;
{  if (bufsize >= 8192) then bufsize := 8191;
  if (bufsize = 0) then bufsize := buffersize;
  data.bufsize := bufsize;}

  cap.version := 0;

  { Set all required hardware capabilities }
  with cap do
  begin
    minfreq         := 4000;
    monoplay.pref   := 22050;
    monoplay.max    := 44100;
    stereoplay.pref := 22050;
    stereoplay.max  := 44100;
    monorecfreq     := 44100;
    stereorecfreq   := 44100;
    flags           := snd_16bit or snd_signed or snd_stereo;
    cardstring      := 'Silent output';
  end;
  cap.driverstring := versionstring;

  getmem(fpustate,4096);

  { end of test }
  data.inchannels  := 0;
  data.outchannels := 0;
  initialized := true;
  chain.next  := nil;
  silentinit  := true;
end;

destructor tsilent.done;
begin
  tobject.done;
  if calls then
  {$IFNDEF FPC}
  setintvec($1C,oldint);
  {$ELSE}
  set_pm_interrupt($1C,oldint);
  {$ENDIF}
  freemem(fpustate,4096);
  initialized := false;
  silentinit  := false;
end;

function tsilent.getcapabilities(var tcap : tcardinfo) : boolean;
begin
  if not initialized then
  begin
    getcapabilities := false;
    exit;
  end else getcapabilities := true;
  tcap := cap;
end;

function outofspec(freq,flags : longint) : boolean;
begin
  outofspec := true;
  if (longint(freq) > cap.monoplay.max)and(cap.flags and snd_stereo = 0) then exit;
  if (flags and snd_stereo <> 0)and(cap.flags and snd_stereo = 0) then exit;
  if (flags and snd_16bit <> 0)and(cap.flags and snd_16bit = 0) then exit;
  if (flags and snd_stereo <> 0)and(freq > cap.stereoplay.max) then exit;
  if (flags and snd_stereo = 0)and(freq > cap.monoplay.max) then exit;
  outofspec := false;
end;

function tsilent.startplayback(freq,flags : longint) : boolean;
begin
  startplayback := false;

  if (flags and snd_auto <> 0) then
  begin
    flags := cap.flags or flags;
    if (cap.flags and snd_stereo = 0) then
    flags := flags and not (snd_stereo or snd_prologic);
  end;

  if (flags and snd_16bit <> 0) then flags := flags or snd_signed;
  if (flags and snd_16bit = 0) then flags := flags and not snd_signed;

  if (flags and snd_prologic <> 0) then flags := flags or snd_stereo;
  data.flags := flags;
  case flags and (snd_stereo or snd_prologic) of
    snd_mono   : begin
                   data.inchannels  := 2;
                   data.outchannels := 1;
                 end;
    snd_stereo : begin
                   data.inchannels  := 2;
                   data.outchannels := 2;
                 end;
    snd_stereo+snd_prologic :
                 begin
                   data.inchannels  := 3;
                   data.outchannels := 2;
                 end;
  end;

  if (freq = 0) then { Set frequency automatically }
  begin
    if (flags and snd_stereo <> 0) then freq := cap.stereoplay.pref
    else freq := cap.monoplay.pref;
  end;

  if outofspec(freq,flags) then exit;
  data.freq := freq;

  data.bufsize := round(freq/18.2);

  data.samplesize := 1;
  if (flags and snd_16bit <> 0) then inc(data.samplesize,data.samplesize);
  if (flags and snd_stereo <> 0) then inc(data.samplesize,data.samplesize);

  getmem(datap,data.bufsize*data.samplesize*2);
  {$IFNDEF FPC}
  getmem32(data.buffer,data.bufsize*data.inchannels*4*2);
  filllong32(ptr48(data.buffer,0),data.bufsize*data.inchannels*2,0);
  {$ELSE}
  getmem(data.buffer,data.bufsize*data.inchannels*4*2);
  filldword(data.buffer^,data.bufsize*data.inchannels*2,0);
  {$ENDIF}

  {$IFNDEF FPC}
  if (data.flags and snd_signed <> 0) then
  begin
    if (data.flags and snd_16bit <> 0) then
      fillword32(ptr48(seg(datap^),ofs(datap^)),data.bufsize*data.samplesize,0) else
      fillchar32(ptr48(seg(datap^),ofs(datap^)),data.bufsize*data.samplesize*2,0);
  end else
  begin
    if (data.flags and snd_16bit <> 0) then
      fillword32(ptr48(seg(datap^),ofs(datap^)),data.bufsize*data.samplesize,32767) else
      fillchar32(ptr48(seg(datap^),ofs(datap^)),data.bufsize*data.samplesize*2,127);
  end;
  {$ELSE}
  if (data.flags and snd_signed <> 0) then
  begin
    if (data.flags and snd_16bit <> 0) then
      fillword(datap^,data.bufsize*data.samplesize,0) else
      fillchar(datap^,data.bufsize*data.samplesize*2,0);
  end else
  begin
    if (data.flags and snd_16bit <> 0) then
      fillword(datap^,data.bufsize*data.samplesize,32767) else
      fillchar(datap^,data.bufsize*data.samplesize*2,127);
  end;
  {$ENDIF}

  callt := addr(self);
  data.freq := freq;
  startplayback := true;
end;

procedure tsilent.stopplayback;
begin
  if not initialized then exit;
  {$IFNDEF FPC}
  freemem32(data.buffer);
  freemem(datap,data.bufsize*data.samplesize*2);
  {$ELSE}
  freemem(data.buffer);
  freemem(datap);
  {$ENDIF}
end;

procedure tsilent.intrcallproc;
var
  i       : longint;
  ch      : psoundchain;
  p,p2,p3 : pointr;

begin
  {$IFNDEF FPC}
  p  := ptr48(seg(datap^),ofs(datap^)+(data.bufsize*data.samplesize)*activeblock);
  p2 := ptr48(data.buffer,(data.bufsize*4*data.inchannels)*activeblock);
  {$ELSE}
  p  := pointer(datap+(data.bufsize*data.samplesize)*activeblock);
  p2 := pointer(data.buffer+(data.bufsize*4*data.inchannels)*activeblock);
  {$ENDIF}

  { Mix the output channels together }
  mixdata(p2);

  {sat32(p2,data.bufsize*data.inchannels);}

  { Convert mixed data to required output format }
  if (data.flags and snd_prologic <> 0) then
  begin
    {$IFNDEF FPC}
    p3 := ptr48(data.buffer,(data.bufsize*4*data.inchannels)*((activeblock+1) and 1));
    {$ELSE}
    p3 := pointer(data.buffer+(data.bufsize*4*data.inchannels)*((activeblock+1) and 1));
    {$ENDIF}
    if (data.flags and snd_signed <> 0) then
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_ds16(p2,p3,p,data.bufsize) else
        aud_s32_to_ds8(p2,p3,p,data.bufsize);
    end else
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_dus16(p2,p3,p,data.bufsize) else
        aud_s32_to_dus8(p2,p3,p,data.bufsize);
    end;
  end else
  if (data.flags and snd_stereo <> 0) then
  begin
    if (data.flags and snd_signed <> 0) then
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_s16(p2,p,data.bufsize*data.samplesize shr 1) else
        aud_s32_to_s8(p2,p,data.bufsize*data.samplesize);
    end else
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_us16(p2,p,data.bufsize*data.samplesize shr 1) else
        aud_s32_to_us8(p2,p,data.bufsize*data.samplesize);
    end;
  end else { We play mono :( }
  begin
    if (data.flags and snd_signed <> 0) then
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_sm16(p2,p,data.bufsize) else
        aud_s32_to_sm8(p2,p,data.bufsize);
    end else
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_usm16(p2,p,data.bufsize) else
        aud_s32_to_usm8(p2,p,data.bufsize);
    end;
  end;
end;

function tsilent.getversionstring : string;
begin
  getversionstring := versionstring;
end;

function tsilent.getnumofoutcontrols : word;
begin
  getnumofoutcontrols := 0;
end;

function tsilent.getoutcontroltype(num : word) : word;
begin
  getoutcontroltype := 0;
end;

function tsilent.getoutcontrolname(cnt : word) : string;
begin
  getoutcontrolname := 'N/A';
end;

function tsilent.getoutcontrolrange(cnt : word) : longint;
begin
  getoutcontrolrange := 0;
end;

function tsilent.getoutcontrol(cnt : word) : longint;
begin
  getoutcontrol := 0;
end;

function tsilent.setoutcontrol(cnt : word;val : longint) : boolean;
begin
  setoutcontrol := false;
end;

begin
  oldexitproc              := exitproc;
  exitproc                 := addr(exitsilent);
end.
