{****************************************************************************
 * SoundLib 2 alpha Version - copyright (c) 1997-2001 by Stefan Goehler     *
 * Copyright agreement in soundlib.txt - illegal copy without this file!    *
 * Any bug reports and suggestions are welcome                              *
 * Complete package and new version always available here:                  *
 * http://sourcenet.home.pages.de                                           *
 ****************************************************************************}
{$IFDEF FPC}

{$A+,B-,F+,G+,N+,E-,P+,Q-,R-,T-,V+,X+}
{$DEFINE NOASM} { Use slow pascal mixing routines }
{$ASMMODE INTEL}
{$ELSE}

{$A+,B-,F+,G+,N+,E-,P+,Q-,R-,T-,V+,X+}
{-$DEFINE NOASM} { Use slow pascal mixing routines }
{$ENDIF}

{$DEFINE VU}     { automatic VU meter calculation on/off     }

{
  History:
   .vu meter calculation modified for better results with centered playback
   .fixed some bugs in MOD-player
   .S3M support added
   .32 bits sample interpolation implemented
   .sample amplification possible by setting volume to 257..512 - overflows
    will be clipped automatically
   .no overflow anymore if sound too loud (e.g. more sounds at once than
    set)
   .sound driver volume control procedures added
   .fixed pause bug - not playing samples were paused and set to play after
    pause without reason
   .sound file unit uses now asm for mixing 16 bits sound data
    -no interpolation support yet
   a3
   .sound muting possible
   .fixed some major32 bits register interferences with GX2 which caused crashes
    -386 registers were not saved (BP still thinks in 286 :\ )
    -FPU operations disturbed MMX operations at the same time
   .added function 'isplaying', 'ispaused'
   .SB playback uses less CPU time
   .added silent driver (drv_sil.pas)
    -use it if no SB found to avoid problems with non-initialized
     objects
    -is able to simulate existing soundcard by retrieving sound data
     from objects (call tsilent.init with true)
    -is able to act as dummy driver - gets commands but doesn't do
     anything (call tsilent.init with false)
   .YEA! FPC-SUPPORT
    After about 10 hours of hanging, checking and changing, SL2 does now
    also support the Free Pascal Compiler. I just want to mention, that this
    compiler - even after years of existing - isn't absolutely bugfree. That
    made me alot of problems.
   .fixed problems with tobject model (object pointers didn't work)
   .whole wave disk playback completely revised
   .added some more comments

}

unit sndlib2;

interface

uses objects
{$IFNDEF FPC}
  ,newfrontier,stream32
{$ENDIF}
  ;

const
  stdbfsize = 1024; { Standard mix chunk size in samples }

  { Volume control constants }
  cnt_main  = $0000; { Main volume       }
  cnt_wave  = $0001; { Wave volume       }
  cnt_cd    = $0002; { CD volume         }
  cnt_midi  = $0003; { MIDI volume       }
  cnt_line  = $0004; { Line in volume    }
  cnt_mic   = $0005; { Microphone vol    }
  cnt_spkr  = $0006; { PC Speaker vol    }

  cnt_agc   = $0007; { AGC control       }
  cnt_gain  = $0008; { Out amplification }

  cnt_trbl  = $0009; { Treble            }
  cnt_bass  = $000A; { Bass              }

  cnt_aux   = $000B; { Auxillary         }
  cnt_spdif = $000C; { SP/DIF            }
  cnt_tad   = $000D; { TAD in            }
  cnt_3d    = $000E; { 3D Audio enhance  }

  cnt_other = $00FF; { Other/unknown control }

  cnt_mask  = $00FF;

  cnt_left  = $0000; { Left channel      }
  cnt_right = $1000; { Right channel     }
  cnt_chn   = $F000;

  cnt_none  = $FFFF;

type
{$IFNDEF FPC}
  pointr  = pointer48;
  dataptr = tselector;
  tstream = tstream32;
  pstream = pstream32;
{$ELSE}
  pointr  = pointer;
  dataptr = pointer;
{$ENDIF}

   tvol = array[0..4] of word;
   tv   = array[0..4] of longint;
   psample = ^tsample;
   psounddat = ^tsounddat;
   psound = ^tsound;

   pdatchain = ^tdatchain;
   tdatchain = record
              current : psounddat;
              next    : pdatchain;
            end;

   psoundchain = ^tsoundchain;
   tsoundchain = record
              current : psound;
              next    : psoundchain;
            end;

   pcardinfo = ^tcardinfo;
   tcardinfo = record
     driverstring   : string[63];
     cardstring     : string[63];
     monoplay : record
       max,pref : longint;
     end;
     stereoplay : record
       max,pref : longint;
     end;
     minfreq        : longint;
     monorecfreq,
     stereorecfreq  : longint;
     flags          : longint;
     version        : longint;
   end;

   psounddriver = ^tsounddriver;
   tsounddriver = object(tobject)
     chain       : tsoundchain;
     initialized : boolean;
     data : record
       listcount   : longint;  { Number of registered tsound objects }
       inchannels,             { Number of input channels (to mix)   }
       outchannels : longint;  { Number of out channels (to card)    }
       samplesize  : longint;  { Size of the output samples in bytes }
       flags       : longint;
       freq        : longint;  { Output frequency                    }
       buffer      : dataptr;  { 32 bit mix buffer selector          }
       bufsize     : longint;  { Buffer size in samples              }
     end;
     constructor init(bufsize : word);
     destructor  done;virtual;

     function    getcapabilities(var tcap : tcardinfo) : boolean;virtual;
     function    startplayback(freq,flags : longint) : boolean;virtual;
     procedure   stopplayback;virtual;
     function    getversionstring : string;virtual;

     function    getnumofoutcontrols : word;virtual;
     function    getoutcontroltype(num : word) : word;virtual;
     function    getoutcontrolname(cnt : word) : string;virtual;
     function    getoutcontrolrange(cnt : word) : longint;virtual;

     function    getoutcontrol(cnt : word) : longint;virtual;
     function    setoutcontrol(cnt : word;val : longint) : boolean;virtual;

     { Out of interest for you :}
     procedure   recalcvol;
     function    get2dvol(chn,vol,posx,posy : longint) : word;virtual;

     procedure   mixdata(p : pointr);
   end;

   tsound = object(tobject)
     output      : psounddriver;
     initialized : boolean;
     bfsize      : word;
     dat : record
       mainfreq    : longint; { Output playback frequency    }
       size        : longint; { Buffer size in bytes         }
       playchnls   : longint; { Maximum channels at once     }
       samplesize  : longint; { Sample size in bytes         }
       samplecount : longint; { Runtime sample count value   }
       lastcount   : longint; { Last samplecount value       }
       listcount   : longint; { Number of registered objects }
       remain      : longint; { Requestblock data remain     }
       vol         : tvol;    { Volume multiplier values     }
     end;

     chain : tdatchain;       { Registered object chain      }

     constructor init(var dest : tsounddriver;channels : longint);
     destructor  done;virtual;

     procedure   setbuffersize(samples : longint);virtual;
     procedure   setchannels(channels : longint);virtual;
     procedure   setvolume(channel,volume : word);virtual;
     function    getvolume(channel : word) : word;virtual;

     function    change : boolean;virtual;

     function    getversionstring : string;virtual;
     procedure   setinterpolation(on : boolean);

     { Out of interest for you :}
     procedure   requestblock(p : pointr;samples : longint);virtual;
     function    getv(chn,vol : word) : longint;virtual;
     procedure   callproc;virtual;
   end;

   tsounddat = object(tobject)
     data : record
       samplesize : longint;   { Sample size in samples            }
       flags      : longint;
       ofreq      : longint;   { Original playback frequency       }
       freq       : longint;   { Current playback frequency        }
       buffer     : dataptr;   { Sound data selector               }
       pos        : longint;   { Current sample pos in sound data  }
       xinc       : longint;   { Position increase delta value     }
       subpos     : longint;
       vol        : tvol;      { Physical channel volumes          }
       v          : tv;        { Logical channel volumes           }
       vuval      : longint;
       vucount    : longint;
       vdelta     : tv;        { Volume increase delta value       }
       vdest      : tv;        { Volume destination value          }
     end;
     father      : psound;
     initialized : boolean;

     constructor init(var dest : tsound);
     destructor  done;virtual;

     function    getnumofsamples : longint;virtual;
     function    getsamplesize : longint;virtual;
     function    getsamplefrequency : longint;virtual;

     procedure   play;virtual;
     procedure   pause;virtual;
     procedure   stop;virtual;
     function    isplaying : boolean;virtual;
     function    ispaused : boolean;virtual;
     procedure   mute(on : boolean);virtual;
     procedure   setvolume(channel,volume : word);virtual;
     function    getvolume(channel : word) : word;virtual;
     procedure   setsoundpos(lr,cs,volume : integer);

     procedure   setpos(pos : longint);virtual;
     function    getpos : longint;virtual;
     function    getsize : longint;virtual;

     procedure   setspeed(speed : double);virtual;
     procedure   setfrequency(freq : longint);virtual;
     procedure   setinterpolation(on : boolean);virtual;
{     procedure   setvolumeramping(on : boolean);}

     function    getvu : byte;

     { Internal usage }
     procedure   requestblock(p : pointr;samples : longint);virtual;
     function    requestblock_direct16(p : pointr;samples : longint) : longint;virtual;
   end;

   tsample = object(tsounddat)
     dat : record
       size           : longint;
       loopstart,stop : longint;
     end;
     constructor init(var dest : tsound;var source : tsounddat);
     destructor  done;virtual;

     procedure   setloop(strt,stp : longint);
     procedure   setpos(pos : longint);virtual;
     function    getpos : longint;virtual;
     function    getsize : longint;virtual;

     procedure   requestblock(p : pointr;samples : longint);virtual;
   end;

   tchannel = object(tsample)
     constructor init(var dest : tsound);
     destructor  done;virtual;

     procedure   setsample(s : dataptr;size,freq : longint);
   end;

const
  snd_command   = $0000000F;
  snd_loopdata  = $000F0000;
  snd_play      = $00000001;
  snd_pause     = $00000002;
  snd_rec       = $00000004;
  snd_mute      = $00000008;

  snd_default   = $00000000;
  snd_mono      = $00000000;
  snd_stereo    = $00000100;
  snd_signed    = $00000200; { Sound data is signed          }
  snd_nonsigned = $00000000; { Sound data is non-signed      }
  snd_8bit      = $00000000; { Sound data uses 8 bits        }
  snd_16bit     = $00000400; { Sound data uses 16 bits       }
  snd_prologic  = $00001000; { Dolby Pro Logic playback      }
  snd_intpol    = $00002000; { use sample interpolation      }
  snd_noramping = $00004000; { don't use volume ramping      }
  snd_loop      = $00010000; { Sample is looped              }
  snd_inloop    = $00080000; { Sample currently in a loop    }

  snd_auto      = $10000000;

{$IFNDEF FPC}
  sl_versionstr    = 'SoundLib2 a3.0213.BP';
{$ELSE}
  sl_versionstr    = 'SoundLib2 a3.0213.FPC';
{$ENDIF}
  sl_versionnum    = $0200;
  sl_versionsubnum = $0213;


implementation

constructor tsounddriver.init(bufsize : word);
begin
  inherited init;
  initialized := true;
end;

destructor tsounddriver.done;
begin
  inherited done;
  initialized := false;
end;

function tsounddriver.getcapabilities(var tcap : tcardinfo) : boolean;
begin
  getcapabilities := false;
end;

function tsounddriver.startplayback(freq,flags : longint) : boolean;
begin
  data.freq  := freq;
  data.flags := flags;
end;

procedure tsounddriver.stopplayback;
begin
end;

procedure tsounddriver.recalcvol;
var
  ch  : psoundchain;
  i,j : integer;

begin
  ch := chain.next;
  for i := 0 to data.listcount-1 do
  begin
    for j := 0 to data.inchannels do
      ch^.current^.setvolume(j,ch^.current^.dat.vol[j]);
    ch := ch^.next;
  end;
end;

function tsounddriver.getversionstring : string;
begin
  getversionstring := '';
end;

function tsounddriver.getnumofoutcontrols : word;
begin
  getnumofoutcontrols := 0;
end;

function tsounddriver.getoutcontroltype(num : word) : word;
begin
  getoutcontroltype := cnt_none;
end;

function tsounddriver.getoutcontrolname(cnt : word) : string;
begin
  getoutcontrolname := 'N/A';
end;

function tsounddriver.getoutcontrolrange(cnt : word) : longint;
begin
  getoutcontrolrange := 0;
end;

function tsounddriver.getoutcontrol(cnt : word) : longint;
begin
  getoutcontrol := 0;
end;

function tsounddriver.setoutcontrol(cnt : word;val : longint) : boolean;
begin
  setoutcontrol := false;
end;

function tsounddriver.get2dvol(chn,vol,posx,posy : longint) : word;
{  Get volume of a channnel in a 2D-field

          X-AXIS

            256



Y
|
A    L       0        R
X  -256              256
I
S


             S
           -256

Position 0 is the listener's position
}
var
  val : longint;

begin
  if (posx < -256) then posx := -256;
  if (posx >  256) then posx :=  256;

  if (posy < -256) then posy := -256;
  if (posy >  256) then posy :=  256;

  case chn of
    0 : val := (((256-posx))*(256-abs(posy))) shr 8;
    1 : val := (((256+posx))*(256-abs(posy))) shr 8;
    2 : begin
          if posy >= 0 then val := 0
          else val := abs(posy)-abs(posx) shr 3;
          if val < 0 then val := 0;
        end;
    else val := 0;
  end;
  if (val > 256) then val := 256;
  get2dvol := val*vol shr 8;
end;

procedure tsounddriver.mixdata(p : pointr);
var
  i       : longint;
  ch      : psoundchain;

begin
  { Reset mixing buffer to 0 }
  {$IFNDEF FPC}
  filllong32(p,data.bufsize*data.inchannels,0);
  {$ELSE}
  filldword(p^,data.bufsize*data.inchannels,0);
  {$ENDIF}

  { Get mix data from all registered tsound objects }
  ch := chain.next;
  for i := 0 to data.listcount-1 do
  begin
    ch^.current^.requestblock(p,data.bufsize);
    ch := ch^.next;
  end;
end;

constructor tsounddat.init(var dest : tsound);
var
  i        : integer;
  newchain : pdatchain;

begin
  inherited init;
  father := addr(dest);
  new(newchain);
  newchain^.next     := father^.chain.next;
  newchain^.current  := addr(self);
  father^.chain.next := newchain;
  inc(father^.dat.listcount);
  data.flags := data.flags or snd_noramping;

  initialized := true;
  for i := 0 to 1 do setvolume(i,256);
  for i := 2 to 4 do setvolume(i,0);
end;

destructor tsounddat.done;
var
  i      : integer;
  ch,ch2 : pdatchain;

begin
  if not initialized then exit;
  ch  := father^.chain.next;
  ch2 := addr(father^.chain);
  for i := 0 to father^.dat.listcount-1 do
  begin
    if (ch^.current = addr(self)) then
    begin
      ch2^.next := ch^.next;
      dispose(ch);
      break;
    end;
    ch2 := ch;
    ch  := ch^.next;
  end;
  dec(father^.dat.listcount);

  inherited done;
  initialized := false;
end;

function tsounddat.getnumofsamples : longint;
begin
  getnumofsamples := 0;
end;

function tsounddat.getsamplesize : longint;
begin
  getsamplesize := data.samplesize;
end;

function tsounddat.getsamplefrequency : longint;
begin
  getsamplefrequency := data.ofreq;
end;

procedure tsounddat.setvolume(channel,volume : word);
begin
  if not initialized then exit;
  if (volume > 512) then volume := 512;
{  if (data.flags and snd_noramping <> 0) then
  begin}
    data.vol[channel] := volume;
    data.v[channel]   := father^.getv(channel,volume);
(*  end else
  begin { Do volume ramping, calculate values }
    data.vdest[channel]  := father^.getv(channel,volume);
    data.vdelta[channel] := longint(data.vdest[channel]-data.v[channel]) div 512;
    data.vol[channel]    := volume;
    {data.v[channel]   := father^.getv(channel,volume);}
  end;*)
end;

function tsounddat.getvolume(channel : word) : word;
begin
  if (channel > father^.output^.data.inchannels) then
    getvolume := 0 else getvolume := data.vol[channel];
end;

procedure tsounddat.setsoundpos(lr,cs,volume : integer);
var
  i : integer;

begin
  if not initialized then exit;
  if (volume > 512) then volume := 512;
  if (volume < 0) then volume := 0;
  for i := 0 to father^.output^.data.inchannels-1 do
    setvolume(i,father^.output^.get2dvol(i,volume,lr,cs));
end;

procedure tsounddat.setpos(pos : longint);
begin
end;

function tsounddat.getpos : longint;
begin
end;

function tsounddat.getsize : longint;
begin
end;


procedure tsounddat.play;
begin
  if not initialized then exit;
  if (data.flags and snd_pause <> 0) then
    data.flags := (data.flags and not snd_pause) or snd_play
  else
  with data do
  begin
    pos    := 0;
    flags  := flags or snd_play and not (snd_inloop);
    subpos := 0;
  end;
end;

procedure tsounddat.pause;
begin
  if not initialized then exit;
  if (data.flags and snd_pause <> 0) then
    data.flags := (data.flags and not snd_pause) or snd_play
  else if (data.flags and snd_play <> 0) then
    data.flags := (data.flags and not snd_play) or snd_pause;
end;

procedure tsounddat.stop;
begin
  if not initialized then exit;
  data.flags := data.flags and not (snd_play or snd_pause or snd_inloop);
end;

function tsounddat.isplaying : boolean;
begin
  isplaying := data.flags and snd_play <> 0;
end;

function tsounddat.ispaused : boolean;
begin
  ispaused := data.flags and snd_pause <> 0;
end;

procedure tsounddat.mute(on : boolean);
begin
  if (on) then
    data.flags := data.flags or (snd_mute)
    else data.flags := data.flags and not (snd_mute);
end;

procedure tsounddat.setspeed(speed : double);
begin
  if not initialized then exit;
  with data do
  begin
    freq := round(ofreq*speed);
    xinc := round(freq/ofreq*65536*ofreq/father^.dat.mainfreq);
  end;
end;

procedure tsounddat.setfrequency(freq : longint);
begin
  if not initialized then exit;
  data.freq := freq;
  with data do
    xinc := round(freq/ofreq*65536*ofreq/father^.dat.mainfreq);
end;

procedure tsounddat.setinterpolation(on : boolean);
begin
  with data do
  if (on) then flags := flags or snd_intpol else flags := flags and not snd_intpol;
end;

{procedure tsounddat.setvolumeramping(on : boolean);
begin
  with data do
  if (on) then flags := flags and not snd_noramping else flags := flags or snd_noramping;
end;}

procedure tsounddat.requestblock(p : pointr;samples : longint);
begin
end;

function tsounddat.requestblock_direct16(p : pointr;samples : longint) : longint;
begin
  requestblock_direct16 := 0;
end;

function tsounddat.getvu : byte;
begin
  if (data.vucount > 0) then
    data.vuval := (data.vuval div data.vucount) div 192 else data.vuval := 0;
  if (data.vuval > 255) then data.vuval := 255;

  getvu := data.vuval;

  data.vuval   := 0;
  data.vucount := 0;
end;

constructor tsound.init(var dest : tsounddriver;channels : longint);
var
  i        : integer;
  newchain : psoundchain;

begin
  inherited init;
  output := addr(dest);
  new(newchain);
  newchain^.next     := output^.chain.next;
  newchain^.current  := addr(self);
  output^.chain.next := newchain;
  inc(output^.data.listcount);
  { Object initialization }
  dat.samplesize  := 2;
  if (output^.data.flags and snd_stereo <> 0) then inc(dat.samplesize,dat.samplesize);

  dat.samplecount := 0;
  dat.remain      := 0;
  dat.mainfreq    := output^.data.freq;
  dat.playchnls   := channels;

  setbuffersize(stdbfsize);

  for i := 0 to output^.data.inchannels do setvolume(i,256);
  chain.next := nil;
  initialized := true;
  output^.recalcvol;
end;

destructor tsound.done;
var
  i      : integer;
  ch,ch2 : psoundchain;

begin
  if not initialized then exit;
  ch  := output^.chain.next;
  ch2 := addr(output^.chain);
  for i := 0 to output^.data.listcount-1 do
  begin
    if (ch^.current = addr(self)) then
    begin
      ch2^.next := ch^.next;
      dispose(ch);
      break;
    end;
    ch2 := ch;
    ch  := ch^.next;
  end;
  dec(output^.data.listcount);

  inherited done;
  initialized := false;
end;

procedure tsound.requestblock(p : pointr;samples : longint);
var
  i          : longint;
  ch         : pdatchain;
  count,size : longint;

begin
  if not initialized then exit;

  count := samples;

  while (count > 0) do
  begin
    if (samples > bfsize) then
    begin
      if (dat.remain > 0) then
      begin
        size := dat.remain;
        dat.remain := 0;
      end
      else
        if (count > bfsize) then size := bfsize else size := count;
    end else
    begin
      if (dat.remain > 0) then
      begin
        if (dat.remain < samples) then
          size := dat.remain else size := samples;
        dec(dat.remain,size);
      end
      else size := count;
    end;

    dec(count,size);
    ch := chain.next;
    for i := 0 to dat.listcount-1 do
    begin
      {$IFDEF FPC}
      if assigned(ch^.current) then
      {$ENDIF}
      ch^.current^.requestblock(p,size);
      ch := ch^.next;
    end;
    inc(dat.samplecount,size);
    if (dat.samplecount mod bfsize = 0) then callproc;

    {$IFNDEF FPC}
    inc(tselofs32(p).offset32,size*output^.data.inchannels*4);
    {$ELSE}
    inc(p,size*output^.data.inchannels*4);
    {$ENDIF}
  end;
  dat.remain := bfsize-dat.samplecount mod bfsize;
end;

procedure tsound.setbuffersize(samples : longint);
begin
  bfsize          := samples;
  dat.size        := bfsize*output^.data.inchannels*4;
  dat.samplecount := 0;
  dat.remain      := 0;
end;

procedure tsound.setchannels(channels : longint);
var
  i : integer;

begin
  dat.playchnls := channels;
  for i := 0 to output^.data.inchannels do setvolume(i,dat.vol[i]);
end;

procedure tsound.setvolume(channel,volume : word);
var
  i  : integer;
  ch : pdatchain;

begin
  if (channel > output^.data.outchannels) then exit;
  if (volume > 512) then volume := 512;
  dat.vol[channel] := volume;
  ch := chain.next;
  for i := 0 to dat.listcount-1 do
  begin
    ch^.current^.setvolume(channel,ch^.current^.data.vol[channel]);
    ch := ch^.next;
  end;
end;

function tsound.getvolume(channel : word) : word;
begin
  if (channel > output^.data.inchannels) then
    getvolume := 0 else getvolume := dat.vol[channel];
end;

function tsound.change : boolean;
begin
  change := (dat.lastcount <> dat.samplecount);
  dat.lastcount := dat.samplecount;
end;

function tsound.getv(chn,vol : word) : longint;
var
  v : longint;
begin
  v := 0;
  if (output^.data.flags and snd_prologic <> 0) then
  begin
(*    if (chn <= 1) then*)
      v := (longint(vol) * dat.vol[chn]) div dat.playchnls div output^.data.listcount div 4
(*    else if chn = 2 then v := round(vol*(dat.vol[chn]/dat.playchnls/father^.data.listcount/4)*0.711); {-3db for S channel}*)
  end else
  begin
    if (chn < output^.data.inchannels) then
      v := (longint(vol) * dat.vol[chn]) div dat.playchnls div output^.data.listcount div 4;
  end;
  getv := v;
end;

procedure tsound.callproc;
begin
end;

function tsound.getversionstring : string;
begin
  getversionstring := sl_versionstr;
end;

procedure tsound.setinterpolation(on : boolean);
var
  ch   : pdatchain;
  i    : integer;

begin
  if not initialized then exit;
  ch := chain.next;
  for i := 0 to dat.listcount-1 do with ch^.current^.data do
  begin
    if (on) then flags := flags or snd_intpol else flags := flags and not snd_intpol;
    ch := ch^.next;
  end;
end;

constructor tsample.init(var dest : tsound;var source : tsounddat);
var
  i : integer;
begin
  inherited init(dest);
  dat.size  := 2*source.getnumofsamples;
  dat.stop  := dat.size shr 1;
  dat.loopstart := $FFFFFFFF;
  data.ofreq := source.getsamplefrequency;
  data.flags := 0;
  data.pos   := 0;
  {$IFNDEF FPC}
  getmem32(data.buffer,dat.size);
  {$ELSE}
  getmem(data.buffer,dat.size);
  {$ENDIF}
  source.play;
  {$IFNDEF FPC}
  source.requestblock_direct16(ptr48(data.buffer,0),source.getnumofsamples);
  {$ELSE}
  source.requestblock_direct16(data.buffer,source.getnumofsamples);
  {$ENDIF}
  setfrequency(data.ofreq);
  data.samplesize := 2;
  for i := 0 to 1 do setvolume(i,256);
  for i := 2 to 4 do setvolume(i,0);
end;

destructor tsample.done;
begin
  if not initialized then exit;
  inherited done;
  {$IFNDEF FPC}
  freemem32(data.buffer);
  {$ELSE}
  freemem(data.buffer);
  {$ENDIF}
end;

procedure tsample.setloop(strt,stp : longint);
begin
  if not initialized then exit;
  if (strt > dat.size shr 1)or(stp < 0)or(stp > dat.size shr 1) then exit;
  if (strt > stp) then exit;
  dat.loopstart := strt;
  data.flags    := data.flags or snd_loop;
  if (stp > 0) then dat.stop := stp else
  begin
    dat.stop      := dat.size shr 1;
    dat.loopstart := $FFFFFFFF;
  end;
end;

procedure tsample.setpos(pos : longint);
begin
  if (not initialized)or(pos < 0) then exit;
  if (pos > dat.stop) then pos := dat.stop;
  data.pos    := pos;
  data.subpos := 0;
  if (pos > dat.loopstart)and(pos < dat.stop) then
  data.flags := data.flags or snd_inloop
  else data.flags := data.flags and not snd_inloop;
end;

function tsample.getpos : longint;
begin
  getpos := data.pos;
end;

function tsample.getsize : longint;
begin
  getsize := dat.size;
end;

procedure tsample.requestblock(p : pointr;samples : longint);
var
  vu       : longint;
  vl       : tv;
  l        : longint;

  {$IFNDEF NOASM}
  stp,stp2 : longint;
  start    : longint;
  xinc     : longint;
  {$ELSE}
  val,val2 : longint;
  chnls,i  : integer;
  {$ENDIF}

begin

  if (not initialized)or(samples = 0) then exit;
  {$IFDEF VU}
  vu    := 0;
  {$ENDIF}
  {$IFNDEF FPC}
  if (data.flags and snd_mute = 0) then
    vl    := data.v
    else fillchar(vl,sizeof(vl),0);
  {$ELSE}
  { Fix FPC bug }
  if (data.flags and snd_mute = 0) then
    for l := 0 to 4 do vl[l] := data.v[l]
    else fillchar(vl,sizeof(vl),0);
  {$ENDIF}
  if (data.flags and snd_play <> 0) then
  begin
    {$IFDEF NOASM}
    {$IFNDEF FPC}
    chnls := father^.father^.data.inchannels-1;
    if (data.flags and snd_intpol <> 0) then
    { Mix sample data interpolated }
    for l := 0 to samples-1 do with data do
    begin
       if (pos >= dat.stop-1) then
       begin
         if {(flags and snd_loop <> 0)}(dat.loopstart <> $FFFFFFFF) then
         begin
           pos := dat.loopstart;
           flags := flags or snd_inloop;
         end
         else   { No looping, stop sample }
         begin
           flags := flags and not snd_play;
           break;
         end;
       end;

       val  := readmem32l(ptr48(buffer,pos shl 1));
       val2 := integer(val shr 16);
       val  := integer(val and $FFFF);
       val  := val+((val2-val)*(subpos shr 1)) div $8000;

       for i := 0 to chnls do
       begin
         if vl[i] > 0 then incmem32l(p,val*vl[i]);
         inc(tselofs32(p).offset32,4);
       end;

       {$IFDEF VU}
       inc(vu,abs(val) shr 8);
       {$ENDIF}

       inc(subpos,xinc);
       inc(pos,subpos shr 16);
       subpos := subpos and $FFFF;
    end
    else { Normal sample data mixing }
    for l := 0 to samples-1 do with data do
    begin
       if (pos >= dat.stop) then
       begin
         if {(flags and snd_loop <> 0)}(dat.loopstart <> $FFFFFFFF) then
         begin
           pos := dat.loopstart;
           flags := flags or snd_inloop;
         end
         else   { No looping, stop sample }
         begin
           flags := flags and not snd_play;
           break;
         end;
       end;

       val := integer(readmem32w(ptr48(buffer,pos shl 1)));

       for i := 0 to chnls do
       begin
         if v[i] > 0 then incmem32l(p,val*vl[i]);
         inc(tselofs32(p).offset32,4);
       end;

       {$IFDEF VU}
       inc(vu,abs(val) shr 8);
       {$ENDIF}

       inc(subpos,xinc);
       inc(pos,subpos shr 16);
       subpos := subpos and $FFFF;
    end;
    {$ELSE} { FPC }
    chnls := father^.output^.data.inchannels-1;
    if (data.flags and snd_intpol <> 0) then
    { Mix sample data interpolated }
    for l := 0 to samples-1 do with data do
    begin
(*       if (pos >= dat.stop)or
          ((flags and snd_inloop <> 0)and(pos < dat.loopstart)) then
       begin
         if {(flags and snd_loop <> 0)}(dat.loopstart <> $FFFFFFFF) then
         begin
           if (flags and snd_loop_pp <> 0) then
           begin
             if (flags and snd_loopback <> 0) then pos := dat.loopstart
             else pos := dat.stop-1;
             flags := flags xor snd_loopback;
           end
           else { Normal looping, restart }
           pos := dat.loopstart;
           flags := flags or snd_inloop;
         end
         else   { No looping, stop sample }
         begin
           flags := flags and not snd_play;
           break;
         end;
       end;*)
       if (pos >= dat.stop-1) then
       begin
         if {(flags and snd_loop <> 0)}(dat.loopstart <> $FFFFFFFF) then
         begin
           pos := dat.loopstart;
           flags := flags or snd_inloop;
         end
         else   { No looping, stop sample }
         begin
           flags := flags and not snd_play;
           break;
         end;
       end;

       val  := longint(pointer(buffer+pos shl 1)^);
       val2 := integer(val shr 16);
       val  := integer(val and $FFFF);

       val  := val+((val2-val)*(subpos shr 1)) div $8000;

       for i := 0 to chnls do
       begin
         if (vl[i] > 0) then inc(longint(p^),val*vl[i]);
         inc(p,4);
       end;

       {$IFDEF VU}
       inc(vu,abs(val) shr 8);
       {$ENDIF}

       inc(subpos,xinc);
       inc(pos,subpos shr 16);
       subpos := subpos and $FFFF;
    end
    else { Normal sample data mixing }
    for l := 0 to samples-1 do with data do
    begin
       if (pos >= dat.stop) then
       begin
         if {(flags and snd_loop <> 0)}(dat.loopstart <> $FFFFFFFF) then
         begin
           pos := dat.loopstart;
           flags := flags or snd_inloop;
         end
         else   { No looping, stop sample }
         begin
           flags := flags and not snd_play;
           break;
         end;
       end;

       val := integer(pointer(buffer+pos shl 1)^);

       for i := 0 to chnls do
       begin
         if (vl[i] > 0) then inc(longint(p^),val*vl[i]);
         inc(p,4);
       end;

       {$IFDEF VU}
       inc(vu,abs(val) shr 8);
       {$ENDIF}

       inc(subpos,xinc);
       inc(pos,subpos shr 16);
       subpos := subpos and $FFFF;
    end;

    {$ENDIF} { FPC }
    {$ELSE} { Use ASM }
    {$IFNDEF FPC}
    begin
      stp   := dat.stop;
      stp2  := dat.stop-1;
      start := dat.loopstart;
      xinc  := data.xinc;
      if (father^.output^.data.inchannels = 2) then
      asm
        db 0Fh,0A8h {push gs}
        db 66h;les   si,p.dword
               mov   ax,es
        db 8Eh,0E8h {mov  gs,ax}
               les   di,self
        db 66h;mov   dx,es:[di].data.flags.word
        db 66h;mov   cx,samples.word
        db 66h;or    cx,cx
               jz    @end_mss
        db 66h;mov   bx,es:[di].data.subpos.word
               mov   ax,es:[di].data.buffer
        db 66h;mov   di,es:[di].data.pos.word
               mov   es,ax

               test  dx,snd_intpol
               jnz   @lpi_mss

      @lp_mss:
        db 66h;cmp   di,stp.word
               jae   @loop_mss
      @cont_mss:

        db 26h,67h,66h,0Fh,0BFh,04h,7Dh,00h,00h,00h,00h { movsx eax,word ptr es:[edi*2]}
        {$IFDEF VU}
        db 66h;push  ax
        {$ENDIF}

        db 66h;push  ax
        db 66h;mul   vl[0].word
        db 65h,66h,67h,01h,06h {add gs:[esi],eax}
        db 66h;pop   ax
        db 66h;mul   vl[4].word
        db 65h,66h,67h,01h,46h,04h {add gs:[esi+4],eax}
        {$IFDEF VU}
        db 66h;pop   ax
        db 66h;or    ax,ax
               jns   @ns
                 db 66h;neg   ax
               @ns:
        db 66h;shr   ax,8
        db 66h;add   vu.word,ax
        {$ENDIF}

        db 66h;add   bx,xinc.word
        db 66h;mov   ax,bx
        db 66h;and   bx,0FFFFh;dw 0
        db 66h;shr   ax,16
        db 66h;add   di,ax

        db 66h;add   si,8
        db 66h;dec   cx
               jnz   @lp_mss
               jmp   @end_mss
      @loop_mss:
        db 66h;cmp   start.word,$FFFF;dw 0FFFFh
               je    @end_mss
        db 66h;mov   di,start.word
               jmp   @cont_mss

               jmp   @end_mss

      @lpi_mss:
        db 66h;cmp   di,stp2.word
               jae   @loopi_mss
        db 26h,67h,66h,8Bh,04h,7Dh,00h,00h,00h,00h { mov eax,es:[edi*2] }
        db 66h,0Fh,0BFh,0D0h  { movsx edx,ax  }
        db 66h;sar   ax,16
      @conti_mss:
        db 66h;sub   ax,dx
        db 66h;push  bx
        db 66h;shr   bx,1
        db 66h,0Fh,0AFh,0C3h  { imul  eax,ebx }
        db 66h;sar   ax,15
        db 66h;add   ax,dx
        db 66h;pop   bx


        {$IFDEF VU}
        db 66h;push  ax
        {$ENDIF}

        db 66h;push  ax
        db 66h;mul   vl[0].word
        db 65h,66h,67h,01h,06h     { add gs:[esi],eax   }
        db 66h;pop   ax
        db 66h;mul   vl[4].word
        db 65h,66h,67h,01h,46h,04h { add gs:[esi+4],eax }
        {$IFDEF VU}
        db 66h;pop   ax
        db 66h;or    ax,ax
               jns   @nsi
                 db 66h;neg   ax
               @nsi:
        db 66h;shr   ax,8
        db 66h;add   vu.word,ax
        {$ENDIF}

        db 66h;add   bx,xinc.word
        db 66h;mov   ax,bx
        db 66h;and   bx,0FFFFh;dw 0
        db 66h;shr   ax,16
        db 66h;add   di,ax

        db 66h;add   si,8
        db 66h;dec   cx
               jnz   @lpi_mss
               jmp   @end_mss
      @loopi_mss:
        db 66h;cmp   start.word,$FFFF;dw 0FFFFh
               je    @end_mss
        db 66h;cmp   di,stp.word
               jae   @restarti_mss
        db 26h,67h,66h,0Fh,0BFh,14h,7Dh,00h,00h,00h,00h { movsx edx,word ptr es:[edi*2] }
        db 66h;push  di
        db 66h;mov   di,start.word
        db 26h,67h,66h,0Fh,0BFh,04h,7Dh,00h,00h,00h,00h { movsx eax,word ptr es:[edi*2] }
        db 66h;pop   di

               jmp   @conti_mss

      @restarti_mss:

        db 66h;mov   di,start.word
               jmp   @lpi_mss

      @end_mss:
        {set pos and subpos}
        les  si,self
        db 66h;mov   es:[si].data.pos.word,di
        db 66h;mov   es:[si].data.subpos.word,bx
        db 0Fh,0A9h {pop gs}
      end else {3 channels}
      asm
        db 0Fh,0A8h {push gs}
        db 66h;les   si,p.dword
               mov   ax,es
        db 8Eh,0E8h {mov  gs,ax}
               les   di,self
        db 66h;mov   dx,es:[di].data.flags.word
        db 66h;mov   cx,samples.word
        db 66h;or    cx,cx
               jz    @end_mss
        db 66h;mov   bx,es:[di].data.subpos.word
               mov   ax,es:[di].data.buffer
        db 66h;mov   di,es:[di].data.pos.word
               mov   es,ax

               test  dx,snd_intpol
               jnz   @lpi_mss

      @lp_mss:
        db 66h;cmp   di,stp.word
               jae   @loop_mss
      @cont_mss:

        db 26h,67h,66h,0Fh,0BFh,04h,7Dh,00h,00h,00h,00h { movsx eax,word ptr es:[edi*2]}
        {$IFDEF VU}
        db 66h;push  ax
        {$ENDIF}
        db 66h;push  ax
        db 66h;mul   vl[0].word
        db 65h,66h,67h,01h,06h {add gs:[esi],eax}
        db 66h;pop   ax
        db 66h;push  ax
        db 66h;mul   vl[4].word
        db 65h,66h,67h,01h,46h,04h {add gs:[esi+4],eax}
        db 66h;pop   ax
        db 66h;mul   vl[8].word
        db 65h,66h,67h,01h,46h,08h {add gs:[esi+8],eax}
        {$IFDEF VU}
        db 66h;pop   ax
        db 66h;or    ax,ax
               jns   @ns
                 db 66h;neg   ax
               @ns:
        db 66h;shr   ax,8
        db 66h;add   vu.word,ax
        {$ENDIF}

        db 66h;add   bx,xinc.word
        db 66h;mov   ax,bx
        db 66h;and   bx,0FFFFh;dw 0
        db 66h;shr   ax,16
        db 66h;add   di,ax

        db 66h;add   si,12
        db 66h;dec   cx
               jnz   @lp_mss
               jmp   @end_mss
      @loop_mss:
        db 66h;cmp   start.word,$FFFF;dw 0FFFFh
               je    @end_mss
        db 66h;mov   di,start.word
               jmp   @cont_mss

               jmp   @end_mss

      @lpi_mss:
        db 66h;cmp   di,stp2.word
               jae   @loopi_mss
        db 26h,67h,66h,8Bh,04h,7Dh,00h,00h,00h,00h { mov eax,es:[edi*2]}
        db 66h,0Fh,0BFh,0D0h  {movsx edx,ax}
        db 66h;sar   ax,16

      @conti_mss:
        db 66h;sub   ax,dx
        db 66h;push  bx
        db 66h;shr   bx,1
        db 66h,0Fh,0AFh,0C3h  {imul  eax,ebx}
        db 66h;sar   ax,15
        db 66h;add   ax,dx
        db 66h;pop   bx

        {$IFDEF VU}
        db 66h;push  ax
        {$ENDIF}
        db 66h;shl   ax,16
        db 66h;sar   ax,16
        db 66h;push  ax
        db 66h;mul   vl[0].word
        db 65h,66h,67h,01h,06h {add gs:[esi],eax}
        db 66h;pop   ax
        db 66h;push  ax
        db 66h;mul   vl[4].word
        db 65h,66h,67h,01h,46h,04h {add gs:[esi+4],eax}
        db 66h;pop   ax
        db 66h;mul   vl[8].word
        db 65h,66h,67h,01h,46h,08h {add gs:[esi+8],eax}
        {$IFDEF VU}
        db 66h;pop   ax
        db 66h;or    ax,ax
               jns   @nsi
                 db 66h;neg   ax
               @nsi:
        db 66h;shr   ax,8
        db 66h;add   vu.word,ax
        {$ENDIF}

        db 66h;add   bx,xinc.word
        db 66h;mov   ax,bx
        db 66h;and   bx,0FFFFh;dw 0
        db 66h;shr   ax,16
        db 66h;add   di,ax

        db 66h;add   si,12
        db 66h;dec   cx
               jnz   @lpi_mss
               jmp   @end_mss
      @loopi_mss:
        db 66h;cmp   start.word,$FFFF;dw 0FFFFh
               je    @end_mss
        db 66h;cmp   di,stp.word
               jae   @restarti_mss
        db 26h,67h,66h,0Fh,0BFh,14h,7Dh,00h,00h,00h,00h { movsx edx,word ptr es:[edi*2]}
        db 66h;push  di
        db 66h;mov   di,start.word
        db 26h,67h,66h,0Fh,0BFh,04h,7Dh,00h,00h,00h,00h { movsx eax,word ptr es:[edi*2]}
        db 66h;pop   di

               jmp   @conti_mss

      @restarti_mss:

        db 66h;mov   di,start.word
               jmp   @lpi_mss

      @end_mss:
        {set pos and subpos}
        les  si,self
        db 66h;mov   es:[si].data.pos.word,di
        db 66h;mov   es:[si].data.subpos.word,bx
        db 0Fh,0A9h {pop gs}
      end;

      if (dat.loopstart = $FFFFFFFF)and(data.pos >= dat.stop) then
        data.flags := data.flags and not snd_play;
    end;
    {$ELSE} { FPC }
    begin
      stp   := dat.stop;
      stp2  := dat.stop-1;
      start := dat.loopstart;
      xinc  := data.xinc;
{      if (father^.father^.data.inchannels = 2) then}
      asm
               mov   esi,p
               mov   edi,self
               mov   eax,[edi].data.flags
               mov   ecx,samples
               or    ecx,ecx
               jz    @end_mss
               mov   ebx,[edi].data.subpos
               mov   edx,[edi].data.buffer
               mov   edi,[edi].data.pos

               test  eax,snd_intpol
               jnz   @lpi_mss

      @lp_mss:
               cmp   edi,stp
               jae   @loop_mss
      @cont_mss:
               movsx eax,word ptr [edi*2+edx]
        {$IFDEF VU}
               push  eax
        {$ENDIF}

               push  eax
               imul  eax,vl[0]
               add   [esi],eax
               pop   eax
               imul  eax,vl[4]
               add   [esi+4],eax
        {$IFDEF VU}
               pop   eax
               or    eax,eax
               jns   @ns
                 neg   eax
               @ns:
               shr   eax,8
               add   vu,eax
        {$ENDIF}

               add   ebx,xinc
               mov   eax,ebx
               and   ebx,0FFFFh
               shr   eax,16
               add   edi,eax

               add   esi,8
               dec   ecx
               jnz   @lp_mss
               jmp   @end_mss
      @loop_mss:
               cmp   start,0FFFFFFFFh
               je    @end_mss
               mov   edi,start
               jmp   @cont_mss

               jmp   @end_mss

      @lpi_mss:
               cmp   edi,stp2
               jae   @loopi_mss
               mov   eax,[edi*2+edx]
               push  edx
               movsx edx,ax
               sar   eax,16
      @conti_mss:
               sub   eax,edx
               push  ebx
               shr   ebx,1
               imul  eax,ebx
               sar   eax,15
               add   eax,edx
               pop   ebx


        {$IFDEF VU}
               push  eax
        {$ENDIF}

               push  eax
               mul   vl[0]
               add   [esi],eax
               pop   eax
               mul   vl[4]
               add   [esi+4],eax
        {$IFDEF VU}
               pop   eax
               or    eax,eax
               jns   @nsi
                 neg   eax
               @nsi:
               shr   eax,8
               add   vu,eax
        {$ENDIF}

               add   ebx,xinc
               mov   eax,ebx
               and   ebx,0FFFFh
               shr   eax,16
               add   edi,eax

               add   esi,8
               dec   ecx
               jnz   @lpi_mss
               jmp   @end_mss
      @loopi_mss:
               cmp   start,0FFFFFFFFh
               je    @end_mss
               cmp   edi,stp
               jae   @restarti_mss
               movsx edx,word ptr [edi*2+edx]
               push  edi
               mov   edi,start
               movsx eax,word ptr [edi*2+edx]
               pop   edi
               pop   edx

               jmp   @conti_mss

      @restarti_mss:

               mov   edi,start
               jmp   @lpi_mss

      @end_mss:
        {set pos and subpos}
               mov   esi,self
               sub   edi,[esi].data.buffer
               mov   [esi].data.subpos,ebx
(*      end else {3 channels}
      asm
               mov   esi,p
               mov   edi,self
               mov   eax,[edi].data.flags
               mov   ecx,samples
               or    ecx,ecx
               jz    @end_mss
               mov   ebx,[edi].data.subpos
               mov   edx,[edi].data.buffer
               mov   edi,[edi].data.pos

               test  eax,snd_intpol
               jnz   @lpi_mss

      @lp_mss:
               cmp   edi,stp
               jae   @loop_mss
      @cont_mss:

               movsx eax,word ptr [edi*2]
        {$IFDEF VU}
               push  eax
        {$ENDIF}
               push  eax
               mul   vl[0]
               add   [esi],eax
               pop   eax
               push  eax
               mul   vl[4]
               add   [esi+4],eax
               pop   eax
               mul   vl[8]
               add   [esi+8],eax
        {$IFDEF VU}
               pop   eax
               or    eax,eax
               jns   @ns
                 neg   eax
               @ns:
               shr   eax,8
               add   vu,ax
        {$ENDIF}

               add   ebx,xinc
               mov   eax,ebx
               and   ebx,0FFFFh
               shr   eax,16
               add   edi,eax

               add   esi,12
               dec   ecx
               jnz   @lp_mss
               jmp   @end_mss
      @loop_mss:
               cmp   start,0FFFFFFFFh
               je    @end_mss
               mov   edi,start
               jmp   @cont_mss

               jmp   @end_mss

      @lpi_mss:
               cmp   edi,stp2
               jae   @loopi_mss
               mov   eax,[edi*2]
               movsx edx,ax
               sar   eax,16

      @conti_mss:
               sub   eax,edx
               push  ebx
               shr   ebx,1
               imul  eax,ebx
               sar   eax,15
               add   eax,edx
               pop   ebx

        {$IFDEF VU}
               push  eax
        {$ENDIF}
               shl   eax,16
               sar   eax,16
               push  eax
               mul   vl[0]
               add   [esi],eax
               pop   eax
               push  eax
               mul   vl[4]
               add   [esi+4],eax
               pop   eax
               mul   vl[8]
               add   [esi+8],eax
        {$IFDEF VU}
               pop   eax
               or    eax,eax
               jns   @nsi
                 neg   eax
               @nsi:
               shr   eax,8
               add   vu,eax
        {$ENDIF}

               add   ebx,xinc
               mov   eax,ebx
               and   ebx,0FFFFh
               shr   eax,16
               add   edi,eax

               add   esi,12
               dec   ecx
               jnz   @lpi_mss
               jmp   @end_mss
      @loopi_mss:
               cmp   start,0FFFFFFFFh
               je    @end_mss
               cmp   edi,stp
               jae   @restarti_mss
               movsx edx,word ptr [edi*2]
               push  edi
               mov   edi,start
               movsx eax,word ptr [edi*2]
               pop   edi

               jmp   @conti_mss

      @restarti_mss:

               mov   edi,start
               jmp   @lpi_mss

      @end_mss:
        {set pos and subpos}
               mov   esi,self
               mov   [esi].data.pos,edi
               mov   [esi].data.subpos,ebx*)
      end;

      if (dat.loopstart = $FFFFFFFF)and(data.pos >= dat.stop) then
        data.flags := data.flags and not snd_play;
    end;
    {$ENDIF} { FPC }
    {$ENDIF} { Use ASM routines }
  end;
  {$IFDEF VU}
  if (data.vol[0] > data.vol[1]) then
    inc(data.vuval,vu*(data.vol[0])) else
    inc(data.vuval,vu*(data.vol[1]));
  inc(data.vucount,samples);
  {$ENDIF}
end;

constructor tchannel.init(var dest : tsound);
begin
  tsounddat.init(dest);
  data.samplesize := 2;
end;

destructor tchannel.done;
begin
  tsounddat.done;
end;

procedure tchannel.setsample(s : dataptr;size,freq : longint);
var
  i : integer;

begin
  data.buffer   := s;
  dat.size      := size;

  data.ofreq    := freq;
  data.flags    := data.flags and not (snd_command or snd_loopdata);
  data.pos      := 0;
  dat.loopstart := $FFFFFFFF;
  dat.stop      := size shr 1;

  setspeed(1);
  for i := 0 to 1 do setvolume(i,256);
  for i := 2 to 4 do setvolume(i,0);
end;

begin
end.

Functions to add:

Functions added:
tsounddat.getvolume
tsounddat.mute
tsound.getvolume
tsound.change
