{****************************************************************************
 * SoundLib 2 sound file reader - copyright (c) 1997-2001 by Stefan Goehler *
 * Copyright agreement in soundlib.txt - illegal copy without this file!    *
 * Complete package and new version always available here:                  *
 * http://sourcenet.home.pages.de                                           *
 ****************************************************************************}

{$IFDEF FPC}
{$A+,B-,F+,G+,N+,E-,P+,Q-,R-,T-,V+,X+}
{$DEFINE NOASM} { Use slow pascal mixing routines    }
{$ASMMODE INTEL}
{$ELSE}

{$A+,B-,F+,G+,N+,E-,P+,Q-,R-,T-,V+,X+}
{$DEFINE NOASM} { ASM routines not complete yet!     }
{$ENDIF}

{$DEFINE VU}    { currently not completely supported }

unit snd_fmt;

interface

{$IFNDEF FPC}
uses sndlib2,stream32,newfrontier,snd_cnv;
{$ELSE}
uses sndlib2,objects,snd_cnv;
{$ENDIF}

type
  pwave = ^twave;
  twave = object(tsounddat)
    dat : record
      datastart  : longint;
      datasize   : longint;
      datapos    : longint;
      databuf    : dataptr;
      load       : longint;
      stream     : pstream;
    end;

    constructor init(var dest : tsound;var s : tstream);
    destructor  done;virtual;

    function    getnumofsamples : longint;virtual;

    procedure   play;virtual;

    procedure   setpos(pos : longint);virtual;
    function    getpos : longint;virtual;
    function    getsize : longint;virtual;

    procedure   requestblock(p : pointr;samples : longint);virtual;
    function    requestblock_direct16(p : pointr;samples : longint) : longint;virtual;
  end;

  praw = ^traw;
  traw = object(twave)
    constructor init(var dest : tsound;var s : tstream;flags,freq : longint);
  end;

  pvoc = ^tvoc;
  tvoc = object(twave)
    constructor init(var dest : tsound;var s : tstream);
  end;

implementation

const
  stdwavebufsize : longint = 8192; { Standard wave buffer size in samples }

constructor twave.init(var dest : tsound;var s : tstream);
type
  twavehdr = packed record {Header of a wavefile}
    formatlength  : longint;
    formattag     : word;
    channels      : word;
    frequency     : longint;
    bytespersecnd : longint;
    blkalign      : word;
    resolution    : word;
  end;

  triff = packed record{20}
    riff      : array[0..3] of char;
    chunksize : longint;
    acon      : array[0..3] of char;
  end;

  tlist = packed record
    list : array[0..3] of char;
    size : longint;
  end;

var
  wavehdr  : twavehdr;
  h        : triff;
  l        : tlist;
  st       : string;
  oldpos   : longint;

  function checkchunks : boolean;
  var
    i        : integer;
  begin
    i := 0;
    checkchunks := false;
    s.read(h,sizeof(h));
    repeat
      s.read(l,sizeof(l));
      if l.list = 'fmt ' then begin
        oldpos := s.getpos;
        s.seek(oldpos-4);
        s.read(wavehdr,sizeof(wavehdr));
        s.seek(oldpos+l.size);
      end
      else if l.list <> 'data' then
      s.seek(s.getpos+l.size);
      inc(i);
    until (l.list = 'data')or(s.status = streaderror)or(i >= 32);
    if i >= 32 then exit;
    if (s.status = streaderror) then exit;
    checkchunks := true;
  end;

begin
  if (s.status <> stok) then fail;
  inherited init(dest);
  data.flags := 0;
  fillchar(dat,sizeof(dat),0);
  dat.stream := @s;
  if not checkchunks then fail;
  dat.datastart := s.getpos;

  if (wavehdr.formattag <> 1) then fail;

  if (wavehdr.resolution <> 8)and(wavehdr.resolution <> 16) then fail;

  data.samplesize := wavehdr.resolution div 8;
  case wavehdr.resolution of
    8  :;
    16 : data.flags := data.flags or snd_16bit;
    else fail;
  end;

  data.ofreq := wavehdr.frequency;

  if wavehdr.bytespersecnd div (data.ofreq*data.samplesize) = 2 then
  begin
    data.flags := data.flags or snd_stereo;
    inc(data.samplesize,data.samplesize);
  end;

  dat.datasize := l.size;

  {$IFNDEF FPC}
  getmem32(data.buffer,stdwavebufsize*4);
  getmem32(dat.databuf,stdwavebufsize*data.samplesize);
  {$ELSE}
  getmem(data.buffer,stdwavebufsize*4);
  getmem(dat.databuf,stdwavebufsize*data.samplesize);
  {$ENDIF}

  dat.datapos := 0;
  setspeed(1);
end;

destructor twave.done;
begin
  if not initialized then exit;
  {$IFNDEF FPC}
  freemem32(data.buffer);
  freemem32(dat.databuf);
  {$ELSE}
  freemem(data.buffer);
  freemem(dat.databuf);
  {$ENDIF}
  inherited done;
end;

function twave.getnumofsamples : longint;
begin
  if not initialized then exit;
  getnumofsamples := dat.datasize div data.samplesize;
end;

procedure twave.play;
begin
  if not initialized then exit;

  if (data.flags and snd_pause = 0) then
  begin
    dat.load := 0;
    dat.stream^.reset;
    dat.stream^.seek(dat.datastart);
    dat.datapos := 0;
    dat.load    := 0;
  end;

  inherited play;
end;

procedure twave.setpos(pos : longint);
begin
  if not initialized then exit;
  if (pos < 0)or(pos*data.samplesize > dat.datasize) then exit;
  dat.stream^.seek(pos*data.samplesize+dat.datastart);
  data.pos    := 0;
  data.subpos := 0;
  dat.load    := 0;
  dat.datapos := pos*data.samplesize;
end;

function twave.getpos : longint;
begin
  if not initialized then exit;
  getpos := (dat.datapos) div data.samplesize-dat.load+data.pos;
end;

function twave.getsize : longint;
begin
  if not initialized then exit;
  getsize := dat.datasize div data.samplesize;
end;

procedure twave.requestblock(p : pointr;samples : longint);
var

  l,i      : longint;
  ssize    : longint;
  val,val2 : longint;
  vl       : tv;

  {$IFNDEF NOASM}
  stp      : longint;
  xinc     : longint;
  count    : longint;
  {$ENDIF}
  vu       : longint;

begin
  if (not initialized)or(samples = 0) then exit;
  {$IFDEF VU}
  vu    := 0;
  {$ENDIF}
  {$IFNDEF FPC}
  if (data.flags and snd_mute = 0) then
    vl    := data.v
    else fillchar(vl,sizeof(vl),0);
  {$ELSE}
  { Fix FPC bug }
  if (data.flags and snd_mute = 0) then
    for l := 0 to 4 do vl[l] := data.v[l]
    else fillchar(vl,sizeof(vl),0);
  {$ENDIF}
  if (data.flags and snd_play <> 0) then
  begin
    if (dat.load = 0) then
      {$IFNDEF FPC}
      dat.load := requestblock_direct16(ptr48(data.buffer,0),stdwavebufsize);
      {$ELSE}
      dat.load := requestblock_direct16(data.buffer,stdwavebufsize);
      {$ENDIF}

    ssize := father^.output^.data.inchannels*4;

    {$IFDEF NOASM}
    {$IFNDEF FPC}
    for l := 0 to samples-1 do with data do
    begin
      if (flags and snd_stereo = 0) then
      begin
        val := integer(readmem32w(ptr48(data.buffer,data.pos shl 1)));
        for i := 0 to father^.output^.data.inchannels-1 do
        incmem32l(ptr48(tselofs32(p).selector,
                  tselofs32(p).offset32+i shl 2),
                  val*vl[i]);
        {$IFDEF VU}
        inc(vu,abs(val) shr 8);
        {$ENDIF}
      end else
      for i := 0 to 1 do
        incmem32l(ptr48(tselofs32(p).selector,
                  tselofs32(p).offset32+i shl 2),
                  longint(integer(readmem32w(ptr48(data.buffer,data.pos shl 2+i shl 1))))*vl[i]);

      inc(subpos,xinc);
      inc(pos,subpos shr 16);
      subpos := subpos and $FFFF;

      inc(tselofs32(p).offset32,ssize);

      if (pos >= dat.load) then
      begin
        if (dat.load < stdwavebufsize) then
        begin
          data.flags := data.flags and not snd_play;
          break;
        end;
        dat.load := requestblock_direct16(ptr48(data.buffer,0),stdwavebufsize);
        pos := pos mod stdwavebufsize;
      end;
    end
    {$ELSE}   { FPC }
    for l := 0 to samples-1 do with data do
    begin
      if (flags and snd_stereo = 0) then
      begin
        val := integer(pointer(data.buffer+data.pos shl 1)^);
        for i := 0 to father^.output^.data.inchannels-1 do
        inc(longint(pointer(p+i shl 2)^),val*data.v[i]);
        {$IFDEF VU}
        inc(vu,abs(val) shr 8);
        {$ENDIF}
      end else
      for i := 0 to 1 do
        inc(longint(pointer(p+i shl 2)^),
            longint(integer(pointer(data.buffer+data.pos shl 2+i shl 1)^))*vl[i]);

      inc(subpos,xinc);
      inc(pos,subpos shr 16);
      subpos := subpos and $FFFF;

      inc(p,ssize);

      if (pos >= dat.load) then
      begin
        if (dat.load < stdwavebufsize) then
        begin
          data.flags := data.flags and not snd_play;
          break;
        end;
        dat.load := requestblock_direct16(data.buffer,stdwavebufsize);
        pos := pos mod stdwavebufsize;
      end;
    end
    {$ENDIF} { FPC }
    {$ELSE}  {NOASM}
    begin
      count := samples;
      xinc  := data.xinc;

      while (count > 0) do
      begin
        asm
          db 0Fh,0A8h {push gs}
          db 66h;les   si,p.dword
                 mov   ax,es
          db 8Eh,0E8h {mov  gs,ax}
                 les   di,self
          db 66h;mov   ax,es:[di].dat.load.word
          db 66h;mov   stp.word,ax
          db 66h;mov   dx,es:[di].data.flags.word
          db 66h;mov   cx,count.word
          db 66h;or    cx,cx
                 jz    @end_ms
          db 66h;mov   bx,es:[di].data.subpos.word
                 mov   ax,es:[di].data.buffer
          db 66h;mov   di,es:[di].data.pos.word
                 mov   es,ax

                 test  dx,snd_stereo
                 jnz   @lp_mss
        @lp_msm:
          db 66h;cmp   di,stp.word
                 jae   @end_ms

          db 26h,67h,66h,0Fh,0BFh,04h,7Dh,00h,00h,00h,00h { movsx eax,word ptr es:[edi*2] }
          {$IFDEF VU}
          db 66h;push  ax
          {$ENDIF}

          db 66h;push  ax
          db 66h;mul   vl[0].word
          db 65h,66h,67h,01h,06h {add gs:[esi],eax}
          db 66h;pop   ax
          db 66h;mul   vl[4].word
          db 65h,66h,67h,01h,46h,04h {add gs:[esi+4],eax}
          {$IFDEF VU}
          db 66h;pop   ax
          db 66h;or    ax,ax
                 jns   @ns_msm
                   db 66h;neg   ax
                 @ns_msm:
          db 66h;shr   ax,8
          db 66h;add   vu.word,ax
          {$ENDIF}

          db 66h;add   bx,xinc.word
          db 66h;mov   ax,bx
          db 66h;and   bx,0FFFFh;dw 0
          db 66h;shr   ax,16
          db 66h;add   di,ax

          db 66h;add   si,ssize.word
          db 66h;dec   cx
                 jnz   @lp_msm

                 jmp   @end_ms

        @lp_mss:
          db 66h;cmp   di,stp.word
                 jae   @end_ms

          db 26h,67h,66h,0Fh,0BFh,04h,0BDh,00h,00h,00h,00h { movsx eax,word ptr es:[edi*4] }
          {$IFDEF VU}
          db 66h;push  ax
          {$ENDIF}

          db 66h;mul   vl[0].word
          db 65h,66h,67h,01h,06h {add gs:[esi],eax}

{          db 26h,67h,66h,0Fh,0BFh,04h,0BDh,00h,00h,00h,00h} { movsx eax,word ptr es:[edi*4] }
          db 26h,67h,66h,0Fh,0BFh,44h,0BDh,02h { movsx eax,word ptr es:[edi*4+2] }

          db 66h;mul   vl[4].word
          db 65h,66h,67h,01h,46h,04h {add gs:[esi+4],eax}
          {$IFDEF VU}
          db 66h;pop   ax
          db 66h;or    ax,ax
                 jns   @ns_mss
                   db 66h;neg   ax
                 @ns_mss:
          db 66h;shr   ax,8
          db 66h;add   vu.word,ax
          {$ENDIF}

          db 66h;add   bx,xinc.word
          db 66h;mov   ax,bx
          db 66h;and   bx,0FFFFh;dw 0
          db 66h;shr   ax,16
          db 66h;add   di,ax

          db 66h;add   si,ssize.word
          db 66h;dec   cx
                 jnz   @lp_mss

        @end_ms:
          db 66h;mov   count.word,cx
          { save pos and subpos }
          les  si,self
          db 66h;mov   es:[si].data.pos.word,di
          db 66h;mov   es:[si].data.subpos.word,bx
          db 0Fh,0A9h {pop gs}
        end;
        if (count > 0) then
        begin
          if (dat.load < stdwavebufsize) then
          begin
            data.flags := data.flags and not snd_play;
            break;
          end;
          dat.load := requestblock_direct16(ptr48(data.buffer,0),stdwavebufsize);
          data.pos := data.pos mod stdwavebufsize;
          inc(tselofs32(p).offset32,(samples-count)*father^.father^.data.inchannels*4);
        end;
      end;
    end;
    {$ENDIF} { NOASM }
  end;
  {$IFDEF VU}
  if (data.vol[0] > data.vol[1]) then
    inc(data.vuval,vu*(data.vol[0])) else
    inc(data.vuval,vu*(data.vol[1]));
  inc(data.vucount,samples);
  {$ENDIF}
end;

function twave.requestblock_direct16(p : pointr;samples : longint) : longint;
var
  i     : integer;
  read  : longint;
  count : longint;
  incr  : longint;
  max   : longint;

begin
  if not initialized then exit;

  if (dat.datapos >= dat.datasize) then
  begin
    requestblock_direct16 := 0;
    exit;
  end;

  count := dat.stream^.getpos;
  if (data.flags and snd_stereo <> 0) then
  incr := 2 else incr := 1;

  { Check for data chunk size overflow }
  if (dat.datapos+samples*data.samplesize > dat.datasize)
  then
  begin
    samples := (dat.datasize-dat.datapos) div data.samplesize;
  end;

  max := samples div stdwavebufsize;
  if samples = stdwavebufsize then max := 0;
  if (data.flags and snd_16bit <> 0) then
  {$IFNDEF FPC}
  dat.stream^.read32(p,samples*data.samplesize)
  {$ELSE}
  dat.stream^.read(p^,samples*data.samplesize)
  {$ENDIF}
  else
  for i := 0 to max do
  begin
    {$IFNDEF FPC}
    read := dat.stream^.read32(ptr48(dat.databuf,0),stdwavebufsize*data.samplesize);
    {$ELSE}
    read := dat.stream^.getpos;
    dat.stream^.read(dat.databuf^,stdwavebufsize*data.samplesize);
    read := dat.stream^.getpos-read;
    {$ENDIF}
    {$IFNDEF FPC}
    if (data.flags and snd_signed <> 0) then
      aud_s8_to_s16(ptr48(dat.databuf,0),p,read div data.samplesize*incr)
      else
      aud_us8_to_s16(ptr48(dat.databuf,0),p,read div data.samplesize*incr);
    inc(tselofs32(p).offset32,stdwavebufsize*2*incr);
    {$ELSE}
    if (data.flags and snd_signed <> 0) then
      aud_s8_to_s16(dat.databuf,p,read div data.samplesize*incr)
      else
      aud_us8_to_s16(dat.databuf,p,read div data.samplesize*incr);
    inc(p,stdwavebufsize*2*incr);
    {$ENDIF}
  end;
  dat.stream^.reset;
  dat.datapos := dat.stream^.getpos;
  count := dat.datapos-count;
  dec(dat.datapos,dat.datastart);
  requestblock_direct16 := count div data.samplesize;
end;


constructor traw.init(var dest : tsound;var s : tstream;flags,freq : longint);
begin
  if (s.status <> stok) then fail;

  tsounddat.init(dest);

  dat.datastart := s.getpos;
  data.flags    := flags;
  data.ofreq    := freq;
  dat.datasize  := s.getsize-dat.datastart;
  dat.stream    := @s;
  data.samplesize := 1;
  if (flags and snd_16bit <> 0) then inc(data.samplesize,data.samplesize);
  if (flags and snd_stereo <> 0) then inc(data.samplesize,data.samplesize);

  {$IFNDEF FPC}
  getmem32(data.buffer,stdwavebufsize*4);
  getmem32(dat.databuf,stdwavebufsize*data.samplesize);
  {$ELSE}
  getmem(data.buffer,stdwavebufsize*4);
  getmem(dat.databuf,stdwavebufsize*data.samplesize);
  {$ENDIF}
  dat.datapos := 0;
  setspeed(1);
end;

constructor tvoc.init(var dest : tsound;var s : tstream);
{I cannot garantuee that this here really works in all cases - this
 seems to be the most stupid file format I've ever seen!
 24 bits integers, time constants instead of Sampling frequencies,
 a lot of needless chunk types...}
type
  tvocheader = packed record
    kennstr    : array[0..$13] of char;
    datastart  : word;
    version    : word;
  end;

  tvocchunk = packed record
    blocktype  : byte;
    blocksize  : array[0..2] of byte; {really stupid! 24 bits integers!!!}
  end;

  tchunk1 = packed record     {preceded by chunktype 8 if any other attributes
                        otherwise the data is mono}
    timeconst  : byte; {really crappy way to save samplingrate...}
    packmethod : byte; {0:8bit; 1:4bit; 2:3bit;3:4bit}
  end;

  tchunk8 = packed record
    timeconst  : word;
    packmethod : byte;
    stereo     : boolean;
  end;

  tchunk9 = packed record
    frequency  : longint;
    resolution : byte;
    channels   : byte;
    format     : word; {0: 8bit unsigned;4:16bit signed}
    reserved   : array[0..3] of byte;
  end;

var
  vochdr    : tvocheader;
  chunk     : tvocchunk;
  chunk1    : tchunk1;
  chunk8    : tchunk8;
  chunk9    : tchunk9;
  blocksize : longint;
  i         : integer;

  function checkchunks : boolean;
  begin
    i := 0;
    checkchunks := false;
    repeat
      s.read(chunk,sizeof(chunk));
      blocksize := longint(chunk.blocksize[2]) shl 16+
                   longint(chunk.blocksize[1]) shl 8+
                   longint(chunk.blocksize[0]);
      inc(i);
      case chunk.blocktype of
      1 : begin
            s.read(chunk1,sizeof(chunk1));
            data.ofreq :=
            256000000 div (65536-(word(chunk1.timeconst) shl 8));
            dat.datasize  := blocksize-sizeof(chunk1);
            if (chunk1.packmethod <> 0) then exit;{ no packing supported }
          end;
      8 : begin
            s.read(chunk8,sizeof(chunk8));
            data.ofreq :=
            256000000 div (65536-((word(chunk8.stereo)+1)*chunk8.timeconst));
            if (chunk8.stereo) then data.flags := data.flags or snd_stereo;
            dat.datasize  := blocksize-sizeof(chunk8);
            if (chunk8.packmethod <> 0) then exit;{ no packing supported }
          end;
      9 : begin
            s.read(chunk9,sizeof(chunk9));
            if (chunk9.channels = 2) then data.flags := data.flags or snd_stereo;
            if (chunk9.format = 4) then data.flags := data.flags or snd_16bit or snd_signed;
            data.ofreq := chunk9.frequency;
            dat.datasize := blocksize-sizeof(chunk9);
          end;
      else s.seek(s.getpos+blocksize);
      end;
    until ((chunk.blocktype = 1)or(chunk.blocktype = 8)or(chunk.blocktype = 9))and
    (blocksize > 12)or(s.status <> stok)or(i >= 32);
    if i >= 32 then exit;
    if (s.status <> stok) then exit;
    checkchunks := true;
  end;

begin
  if (s.status <> stok) then fail;

  tsounddat.init(dest);
  data.flags := 0;

  s.read(vochdr,sizeof(vochdr));
  if vochdr.kennstr <> 'Creative Voice File'#26 then fail;
  s.seek(vochdr.datastart);
  if not checkchunks then fail;
  dat.datastart := s.getpos;
  dat.datasize  := s.getsize-dat.datastart;

  dat.datastart := vochdr.datastart;

  data.samplesize := 1;
  if (data.flags and snd_16bit <> 0) then inc(data.samplesize,data.samplesize);
  if (data.flags and snd_stereo <> 0) then inc(data.samplesize,data.samplesize);

  dat.stream := @s;

  {$IFNDEF FPC}
  getmem32(data.buffer,stdwavebufsize*4);
  getmem32(dat.databuf,stdwavebufsize*data.samplesize);
  {$ELSE}
  getmem(data.buffer,stdwavebufsize*4);
  getmem(dat.databuf,stdwavebufsize*data.samplesize);
  {$ENDIF}
  dat.datapos := 0;
  setspeed(1);
end;

begin
end.
